local addonName, ns = ...
local _G = _G
local addon = ns.addon

local barTemplate = addon.barTemplate
addon.xpBarTemplate = addon:newTemplate(barTemplate)
local xpBarTemplate = addon.xpBarTemplate
local setfenv = _G.setfenv
setfenv(1, xpBarTemplate)

local min, max, floor, ceil = _G.min, _G.max, _G.floor, _G.ceil
local log10 = _G.log10
local type, pairs, ipairs  = _G.type, _G.pairs, _G.ipairs
local assert, error = _G.assert, _G.error
local string, tostring = _G.string, _G.tostring

local MIN_WIDTH = 70
local MAX_DIGITS = 5


xpBarTemplate.defaults.startXP = 0
xpBarTemplate.defaults.targetXP = nil

xpBarTemplate.optionsTable.args.range = {
    name = "Range",
    desc = "Start and target values.",
    type = 'group',
    args = {
        startXP = {
            name = "Start XP",
            desc = "Minimum XP value to display.",
            type = 'range',
            min = 0,
            max = nil,
            step = 1,
            bigStep = nil,
            get = 'getStartXP',
            set = 'setStartXP',
        },
        startLevel = {
            name = "Start Level",
            desc = "Minimum level to display.",
            type = 'range',
            min = 1,
            max = nil,
            bigStep = 1,
            get = 'getStartLevel',
            set = 'setStartLevel',
        },
        targetXP = {
            name = "Target XP",
            desc = "Maximum XP value to display.",
            type = 'range',
            min = nil,
            max = nil,
            step = 1,
            bigStep = nil,
            get = 'getTargetXP',
            set = 'setTargetXP',
        },
        targetLevel = {
            name = "Target Level",
            desc = "Maximum level to display.",
            type = 'range',
            min = nil,
            max = nil,
            bigStep = 1,
            get = 'getTargetLevel',
            set = 'setTargetLevel',
        },
    },
}

-- start xp
function xpBarTemplate:getStartXP(info)
    local startXP = self:getOptions().startXP or self.defaults.startXP
    local startXPMax = self:getStartXPMax()
    assert(startXP, "startXP missing")
    assert(startXPMax, "startXPMax missin")
    if startXP > startXPMax then
        return startXPMax
    else
        return startXP
    end
end

function xpBarTemplate:getStartXPMin()
    error("Abstract function getStartXPMin defined in xpBarTemplate must be implemented.")
end

function xpBarTemplate:getStartXPMax()
    return self:getCurrentXP()
end

function xpBarTemplate:setStartXP(info,value)
    local startXPMin = self:getStartXPMin()
    local startXPMax = self:getStartXPMax()
    if value < startXPMin then
        value = startXPMin
    end
    if value > startXPMax then
        value = startXPMax
    end
    self:getOptions().startXP = value
    self:Invalidate()
end

-- start level
function xpBarTemplate:getStartLevel(info)
    local startXP = self:getStartXP()
    local level = self:calcLevelForXP(startXP)
    return level
end

function xpBarTemplate:getStartLevelMin()
    local startXPMin = self:getStartXPMin()
    local levelMin = self:calcLevelForXP(startXPMin)
    return levelMin
end

function xpBarTemplate:getStartLevelMax()
    local startXPMax = self:getStartXPMax()
    local levelMax = self:calcLevelForXP(startXPMax)
    return levelMax
end

function xpBarTemplate:setStartLevel(info,value)
    local prevLevel = floor(value)
    local prevLevelXP = self:calcXpForLevel(prevLevel)
    local nextLevel = ceil(value)
    if prevLevel < nextLevel then
        local nextLevelXP = self:calcXpForLevel(nextLevel)
        local startXP =  prevLevelXP + (value-prevLevel)*(nextLevelXP-prevLevelXP)
        self:setStartXP(info,startXP)
    else
        self:setStartXP(info,prevLevelXP)
    end
end

-- target xp
function xpBarTemplate:getTargetXP(info)
    return self:getOptions().targetXP or self:getTargetXPMax()
end

function xpBarTemplate:getTargetXPMin()
    return self:getCurrentXP()
end

function xpBarTemplate:getTargetXPMax()
    local maxTargetLevel = self:getTargetLevelMax()
    return self:calcXpForLevel(maxTargetLevel)
end

function xpBarTemplate:setTargetXP(info,value)
    self:getOptions().targetXP = value
    self:Invalidate()
end

-- target level
function xpBarTemplate:getTargetLevel(info)
    local targetXP = self:getTargetXP()
    local level = self:calcLevelForXP(targetXP)
    return level
end

function xpBarTemplate:getTargetLevelMin()
    local targetXPMin = self:getTargetXPMin()
    local levelMin = self:calcLevelForXP(targetXPMin)
    return levelMin
end

function xpBarTemplate:getTargetLevelMax()
    error("Abstract function getTargetLevelMax defined in xpBarTemplate must be implemented.")
end

function xpBarTemplate:setTargetLevel(info,value)
    assert(self and self.optionsTable and self.optionsTable.name, "where is my optionsTable?")
    assert(value and type(value) == 'number',"suplied argument value is not a number: name=" .. tostring(self.optionsTable.name) .. ", value=" .. tostring(value))
    local prevLevel = floor(value)
    local prevLevelXP = self:calcXpForLevel(prevLevel)
    local nextLevel = ceil(value)
    assert(prevLevel,"prev level is nil. name=" .. tostring(self.optionsTable.name) .. " value=" .. tostring(value))
    assert(nextLevel,"next level is nil. name=" .. tostring(self.optionsTable.name) .. " value=" .. tostring(value))
    if prevLevel < nextLevel then
        local nextLevelXP = self:calcXpForLevel(nextLevel)
        local targetXP =  prevLevelXP + (value-prevLevel)*(nextLevelXP-prevLevelXP)
        self:setTargetXP(info,targetXP)
    else
        self:setTargetXP(info,prevLevelXP)
    end
end

-- end of setters and getters

function xpBarTemplate:OnInitialize()
    barTemplate.OnInitialize(self)
    self:updateOptionsTable()
end

function xpBarTemplate:OnEnable()
    barTemplate.OnEnable(self)
    self:updateOptionsTable()
end

function xpBarTemplate:updateOptionsTable()
    --update options table
    local startXPMin = self:getStartXPMin()
    local startXPMax = self:getStartXPMax()
    local oTA = self.optionsTable.args.range.args
    oTA.startXP.min = startXPMin
    oTA.startXP.max = startXPMax
    oTA.startLevel.max  = self:getStartLevelMax()
    oTA.targetXP.max    = self:getTargetXPMax()
    oTA.targetLevel.max = self:getTargetLevelMax()
    oTA.targetXP.min    = oTA.startXP.max
    oTA.targetLevel.min = oTA.startLevel.max
    local targetXPMin = oTA.targetXP.min
    local targetXPMax = oTA.targetXP.max
    assert(startXPMin <= startXPMax, "startXPMin=" .. tostring(startXPMin)..", startXPMax="..tostring(startXPMax))
    assert(oTA.startLevel.min  <= oTA.startLevel.max)
    assert(targetXPMin<=targetXPMax, "targetXPMin="..tostring(targetXPMin)..", targetXPMax="..tostring(targetXPMax))
    assert(oTA.targetLevel.min <= oTA.targetLevel.max)
end

function xpBarTemplate:getCurrentXP()
    error("Abstract function getCurrentXP defined in xpBarTemplate must be implemented.")
end

function xpBarTemplate:getCurrentLevel()
    return self:calcLevelForXP(self:getCurrentXP())
end



local function startNumbers(cache)
    cache.textures = cache.textures or {}
    cache.strings = cache.strings or {}
    cache.visible = 0
end

local function finishNumbers(cache)
    for i,texture in ipairs(cache.textures) do
        if i >= cache.visible then
            texture:Hide()
        end
    end
    for i,string in ipairs(cache.strings) do
        if i >= cache.visible then
            string:Hide()
       end
    end
end


function xpBarTemplate:levelToStringAndSize(level,digits,baseFontSize)
    local strLevel = string.format('%.'..digits..'f',level)
    local strLevelShort = strLevel
    if string.find(strLevel, '%.') then
        strLevelShort = string.gsub(strLevel, '0+$', '')
    end
    -- calculate font size
    local realDigits = digits - string.len(strLevel) + string.len(strLevelShort)
    --addon:Debug(level,digits)
    local size = max(baseFontSize - realDigits * 1,4)
    local strLevelClean = string.gsub(strLevelShort,'%.$','')
    return strLevelClean, size
end

--
function xpBarTemplate:drawElement(cache, frame, flow, position, level, baseFontSize, digits)
    local index = cache.visible
    -- texture
    local levelTexture = cache.textures[index]
    if not levelTexture then
        levelTexture = frame:CreateTexture(nil,'OVERLAY')
        cache.textures[index] = levelTexture
        levelTexture:SetTexture(1,1,1,1)
    end
    levelTexture:ClearAllPoints()
    if flow == 'LR' then
        levelTexture:SetWidth(1)
        levelTexture:SetPoint('TOP')
        levelTexture:SetPoint('BOTTOM')
        levelTexture:SetPoint('LEFT',frame,'LEFT', position,0)
    elseif flow == 'RL' then
        levelTexture:SetWidth(1)
        levelTexture:SetPoint('TOP')
        levelTexture:SetPoint('BOTTOM')
        levelTexture:SetPoint('RIGHT',frame,'RIGHT', -position,0)
    elseif flow == 'TB' then
        levelTexture:SetHeight(1)
        levelTexture:SetPoint('LEFT')
        levelTexture:SetPoint('RIGHT')
        levelTexture:SetPoint('TOP',frame,'TOP', 0, -position)
    elseif flow == 'BT' then
        levelTexture:SetHeight(1)
        levelTexture:SetPoint('LEFT')
        levelTexture:SetPoint('RIGHT')
        levelTexture:SetPoint('BOTTOM',frame,'BOTTOM', 0, position)
    else
        error("Invalid flow")
    end

    levelTexture:Show()

    -- string
    local levelString = cache.strings[index]
    if not levelString then
        levelString = frame:CreateFontString(nil,'OVERLAY')
        cache.strings[index] = levelString
    end
    local strLevelClean, size = self:levelToStringAndSize(level,digits,baseFontSize)
    local isValid = levelString:SetFont('Fonts\\FRIZQT__.TTF',size)
    assert(isValid)
    levelString:SetText(strLevelClean)

    levelString:ClearAllPoints()
    if flow == 'LR' then
        levelString:SetJustifyV('MIDDLE')
        if position > 20 then
            levelString:SetPoint('RIGHT',levelTexture,'LEFT')
            levelString:SetJustifyH('RIGHT')
        else
            levelString:SetPoint('LEFT',levelTexture,'RIGHT')
            levelString:SetJustifyH('LEFT')
        end
    elseif flow == 'RL' then
        levelString:SetJustifyV('MIDDLE')
        if position > 20 then
            levelString:SetPoint('LEFT',levelTexture,'RIGHT')
            levelString:SetJustifyH('LEFT')
        else
            levelString:SetPoint('RIGHT',levelTexture,'LEFT')
            levelString:SetJustifyH('RIGHT')
        end
    elseif flow == 'TB' then
        levelString:SetJustifyH('CENTER')
        if position > 20 then
            levelString:SetPoint('BOTTOM',levelTexture,'TOP')
            levelString:SetJustifyV('BOTTOM')
        else
            levelString:SetPoint('TOP',levelTexture,'BOTTOM')
            levelString:SetJustifyV('TOP')
        end
    elseif flow == 'BT' then
        levelString:SetJustifyH('CENTER')
        if position > 20 then
            levelString:SetPoint('TOP',levelTexture,'BOTTOM')
            levelString:SetJustifyV('TOP')
        else
            levelString:SetPoint('BOTTOM',levelTexture,'TOP')
            levelString:SetJustifyV('BOTTOM')
        end
    end
    levelString:Show()
    cache.visible = index + 1
end



-- start is inclusive, end is exclusive
function xpBarTemplate:drawNumbersSub(cache, frame, flow, length, fontSize, startPos, startLevel, endPos, endLevel, digits)
    --if length < MIN_WIDTH then
    --    return
    --end

    if endPos < 0 or length < startPos then
        return
    end
    --addon:Debug("TRACE: startPos=" .. startPos ..
    --            ", startLevel=" .. startLevel ..
    --             ", endPos=" .. endPos ..
    --            ", endLevel=" .. endLevel ..
    --            ", width=" .. (endPos-startPos))

    if startLevel < endLevel then
        local currentLength = endPos - startPos
        -- enough space for 10 sub levels
        if (currentLength >= (10 * MIN_WIDTH)) and digits < MAX_DIGITS then
            local stepSize = (endLevel-startLevel)/10
            if stepSize <= 0 then
                return
            end
            for i = 0,9 do
                local subStartLevel = startLevel +   i   * stepSize
                local subEndLevel   = startLevel + (i+1) * stepSize
                local subStartPos = startPos +   i   / 10 * currentLength
                local subEndPos   = min(startPos + (i+1) / 10 * currentLength,endPos)
                assert(subEndPos-subStartPos < currentLength)
                assert(subStartPos >= startPos)
                assert(subEndPos <= endPos, "i=" .. i ..
                                            ", subEndPos=" .. subEndPos .. ", endPos=" .. endPos ..
                                            ", delta=" .. (subEndPos-endPos))
                assert(subEndLevel-subStartLevel < endLevel-startLevel)
                assert(subStartLevel >= startLevel)
                assert(subEndLevel <= endLevel)
                self:drawNumbersSub(cache, frame, flow, length, fontSize,
                               subStartPos, subStartLevel, subEndPos, subEndLevel, digits+1)
            end
        else
            -- not enough space for 10 sublevels,
            self:drawElement(cache, frame, flow, startPos, startLevel, fontSize, digits)
        end
    else
        -- single element
        if 0 <= endPos and endPos < length then
            self:drawElement(cache, frame, flow, endPos, endLevel, fontSize, digits)
        end
    end
end

function xpBarTemplate:drawNumbers()
    if ns.debugDrawNumbers then
        addon:Debug(self:GetName() .. " in xpBarTemplate:drawNumbers()")
    end
    local startLevel = floor(self:getStartLevel())
    local targetLevel = ceil(self:getTargetLevel())
    assert(type(startLevel) == 'number', "startLevel is a " .. type(startLevel))
    local options = self:getOptions()
    local barCache = self.cache
    barCache.levels = barCache.levels or {}
    local levelCache = barCache.levels

    local frame = barCache.frame

    startNumbers(levelCache)

    if not (startLevel < targetLevel) then
        finishNumbers(levelCache)
        return
    end

    if (not frame:GetRight()) or (not frame:GetLeft()) then
        addon:Debug("Frame size not available")
        finishNumbers(levelCache)
        return
    end
    local width = frame:GetRight() - frame:GetLeft()
    local height = frame:GetTop() - frame:GetBottom()
    local flow = self:getOptions().flow or self.defaults.flow
    local horizontal = flow == 'LR' or flow == 'RL'
    local length = horizontal and width or height
    if (length < MIN_WIDTH) then
        finishNumbers(levelCache)
         if ns.debugDrawNumbers then
            addon:Debug(self:GetName() .. " in xpBarTemplate:drawNumbers(): length < MIN_WIDTH",width,height)
        end
        return
    end

    local fontSize = horizontal and height or width

    local startGlobalXP = self:getStartXP()
    local targetGlobalXP = self:getTargetXP()
    local deltaXP = targetGlobalXP - startGlobalXP
    if deltaXP > 0 then
        -- loop over full levels
        for level = startLevel,targetLevel do
            local startXpForLevel = self:calcXpForLevel(level)
            local endXpForLevel
            if level < targetLevel then
                endXpForLevel = self:calcXpForLevel(level+1)
                --addon:Debug("level < targetLevel", level, targetLevel)
            else
                endXpForLevel = startXpForLevel
                --addon:Debug("level >= targetLevel", level, targetLevel, self:calcXpForLevel(level+1), startXpForLevel, targetGlobalXP)
            end
            local startRatio = (startXpForLevel-startGlobalXP)/deltaXP
            local endRatio =   (endXpForLevel  -startGlobalXP)/deltaXP
            local curPos = startRatio * length
            local nextPos = endRatio * length
            self:drawNumbersSub(levelCache, frame, flow, length, fontSize, curPos, level, nextPos, level+1, 0)
        end
    end

    finishNumbers(levelCache)
end