local addonName, ns = ...
local _G = _G
local addon = ns.addon
local moduleTemplate = addon.moduleTemplate
addon.barTemplate = addon:newTemplate(moduleTemplate)
local barTemplate = addon.barTemplate
local setfenv = _G.setfenv
setfenv(1, barTemplate)

local assert, error = _G.assert, _G.error
local type = _G.type
local max = _G.max
local string, tostring = _G.string, _G.tostring

local UIParent = _G.UIParent
local CreateFrame = _G.CreateFrame

local Locales = ns.Locales

local SidePanel
do
    local LibStub = _G.LibStub
    SidePanel = LibStub:GetLibrary('LibSidePanel-1.0', false)
end

local AUTOMATIC = 'AUTOMATIC'
local MANUAL = 'MANUAL'

--[==[
-- Common settings and options for modules
--]==]

barTemplate.optionsTable.args.layout = {
    name = "Layout",
    desc = "Layout of the bar, including position and size.",
    type = 'group',
    get = 'genericGet',
    set = 'layoutSet',
    args = {
        layoutType = {
            name = "Layout type",
            desc = "Type of layout. Either automatic or manual.",
            type = 'select',
            values = { MANUAL = 'manual' , AUTOMATIC = 'automatic' },
        },
        size = {
            name = "Size",
            desc = "Size (i.e, width or height) of the bar.",
            type = 'range',
            min = 1,
            max = 50,
            bigStep = 1,
            disabled = 'isManualLayout',
        },
        priority = {
            name = "priority",
            desc = "Priority of the bar. Lower values move the frame outwards.",
            type = 'range',
            min = 1,
            max = 200,
            bigStep = 1,
            disabled = 'isManualLayout',
        },
        side = {
            name = "Side",
            desc = "Side of the screen to place the bar.",
            type = 'select',
            values = { TOP = 'TOP', LEFT = 'LEFT', RIGHT = 'RIGHT', BOTTOM = 'BOTTOM' },
            disabled = 'isManualLayout',
        },
        top = {
            name = "Top",
            desc = "Distance of top edge from the lower screen edge.",
            type = 'range',
            min = nil,
            max = nil,
            bigStep = 1,
            disabled = 'isAutomaticLayout',
        },
        bottom = {
            name = "Bottom",
            desc = "Distance of bottom edge from the lower screen edge.",
            type = 'range',
            min = 0,
            max = nil,
            bigStep = 1,
            disabled = 'isAutomaticLayout',
        },
        left = {
            name = "Left",
            desc = "Distance of left edge from the left screen edge.",
            type = 'range',
            min = 0,
            max = nil,
            bigStep = 1,
            disabled = 'isAutomaticLayout',
        },
        right = {
            name = "Right",
            desc = "Distance of right edge from the left screen edge.",
            type = 'range',
            min = nil,
            max = nil,
            bigStep = 1,
            disabled = 'isAutomaticLayout',
        },
        flow = {
            name = "Flow",
            desc = "Flow of the bar.",
            type = 'select',
            values = { LR = "Left to right",
                       RL = "Right to left",
                       TB = "Top to bottom",
                       BT = "Bottom to top", },
        },
        priColor = {
            name = "Primary Color",
            type = 'color',
            get = 'colorGetter',
            set = 'colorSetter',
        },
        secColor = {
            name = "Secondary Color",
            type = 'color',
            get = 'colorGetter',
            set = 'colorSetter',
        },
        backColor = {
            name = "Background Color",
            type = 'color',
            get = 'colorGetter',
            set = 'backColorSetter',
        },
    }, -- layout.args
} -- layout

--[=[
-- type, size, priority, and side are placement options, they trigger updateBarPlacement
--]=]
function barTemplate:genericGet(info)
    local key = info[#info]
    return self:getOptions()[key] or self.defaults[key]
end

function barTemplate:layoutSet(info,value)
    local key = info[#info]
    self:getOptions()[key] = value
    self:updateBarPlacement()
end

function barTemplate:colorGetter(info)
    local key = info[#info]
    local color = self:getOptions()[key] or self.defaults[key]
    return color.r, color.g, color.b, color.a
end

function barTemplate:colorSetter(info,r,g,b,a)
    local key = info[#info]
    local color = self:getOptions()[key] or self.defaults[key]
    color.r = r
    color.g = g
    color.b = b
    color.a = a
    self:Invalidate()
end

function barTemplate:backColorSetter(info,r,g,b,a)
    local key = info[#info]
    local color = self:getOptions()[key]
    if not color then
        color = {}
        self:getOptions()[key] = color
    end
    color.r = r
    color.g = g
    color.b = b
    color.a = a
    local cache = self.cache
    if cache then
        local frame = cache.frame
        if frame then
            local texture = frame.texture
            texture:SetTexture(r,g,b,a)
        end
    end
end

function barTemplate:isManualLayout(info)
    local layoutType = self:getOptions().layoutType or self.defaults.layoutType
    return layoutType == MANUAL
end

function barTemplate:isAutomaticLayout(info)
    local layoutType = self:getOptions().layoutType or self.defaults.layoutType
    return layoutType == AUTOMATIC
end

function barTemplate:getTopMax(info)
    return UIParent:GetHeight()
end

function barTemplate:getRightMax(info)
    return UIParent:GetWidth()
end

--[=[
-- Default values for the options.
--]=]
barTemplate.defaults.layoutType = AUTOMATIC
barTemplate.defaults.side = 'BOTTOM'
barTemplate.defaults.size = 10
barTemplate.defaults.priority = 100
barTemplate.defaults.flow = 'LR'
barTemplate.defaults.priColor= { r = 0, g = 0, b = 1, a = 1 }
barTemplate.defaults.secColor= { r = 0, g = 0.5, b = 0, a = 1 }
barTemplate.defaults.backColor= { r = 0, g = 0, b = 0, a = 1 }

function barTemplate:isEnabled()
    return self:getOptions().enable
end

--- Called when the addon is enabled.
function barTemplate:OnEnable()
    if ns.debugOnEnDisable then
        addon:Debug(self:GetName() .. " in barTemplate:OnEnable()")
    end
    if self:isEnabled() then
        self:createBar()
    else
        self:Disable()
    end

    self.optionsTable.args.layout.args.top.max = self:getTopMax()
    self.optionsTable.args.layout.args.right.max = self:getRightMax()
end

--- Called when the addon is disabled
function barTemplate:OnDisable()
    if ns.debugOnEnDisable then
        addon:Debug(self:GetName() .. " in barTemplate:OnDisable()")
    end
    self:destroyBar()
end

--- Creates a new bar or enables an existing one.
function barTemplate:createBar()
    if ns.debugOnEnDisable then
        addon:Debug(self:GetName() .. " in barTemplate:createBar()")
    end
    local barOptions = self:getOptions()
    assert(barOptions)
    self.cache = self.cache or {}
    local barCache = self.cache
    assert(barCache)

    barCache.frame = barCache.frame or CreateFrame('Frame', addonName .. self:GetName() .. 'Bar', UIParent)
    assert(barCache.frame)
    local frame = barCache.frame
    -- create a texture
    frame.texture = frame.texture or frame:CreateTexture(nil, 'BACKGROUND')
    local texture = frame.texture
    texture:ClearAllPoints()
    texture:SetPoint('TOPLEFT')
    texture:SetPoint('BOTTOMRIGHT')
    local c = barOptions.backColor or self.defaults.backColor
    texture:SetTexture(c.r,c.g,c.b,c.a)
    frame:Show()
    -- Hook some scripts
    if not self:IsHooked(frame, 'OnSizeChanged') then
        self:HookScript(frame, 'OnSizeChanged')
    end
    if not self:IsHooked(frame, 'OnEnter') then
        self:HookScript(frame, 'OnEnter')
    end
    if not self:IsHooked(frame, 'OnLeave') then
        self:HookScript(frame, 'OnLeave')
    end

    self:updateBarPlacement()
end

function barTemplate:OnSizeChanged(event)
    self:Invalidate(event)
end

--- Invalidates the bar
function barTemplate:Invalidate(event)
    --self.invalid = true
    local frame = self.cache.frame
    if frame and not self:IsHooked(frame, 'OnUpdate') then
        self:HookScript(frame,'OnUpdate','OnUpdate')
    end
end

function barTemplate:OnUpdate()
    self:Unhook(self.cache.frame,'OnUpdate')
    --if not self.invalid then
    --    return
    --end
    --self.invalid = nil
    --addon:Debug("Redraw")
    self:Redraw()
end

--- Updates the bar placement. This includes setting the width and adding to LibSidePanel.
function barTemplate:updateBarPlacement()
    local barOptions = self:getOptions()
    assert(barOptions)
    local layoutType = barOptions.layoutType or self.defaults.layoutType
    if layoutType == AUTOMATIC then
        self:updateBarPlacementAutomatic()
    elseif layoutType == MANUAL then
        self:updateBarPlacementManual()
    end
end

function barTemplate:updateBarPlacementAutomatic()
    local barOptions = self:getOptions()
    local barCache = self.cache
    assert(barCache, "Where is my cache?")
    if self:IsEnabled() then
        --addon:Debug("UpdatePlacement")
        local frame = barCache.frame
        assert(frame, "Where is my frame?")
        local size = barOptions.size or self.defaults.size
        assert(type(size)=='number',"Size is not a number, but a " .. type(size))
        local side = barOptions.side or self.defaults.side
        if side == 'TOP' or side == 'BOTTOM' then
            frame:SetHeight(size)
        else
            frame:SetWidth(size)
        end
        SidePanel:Register(frame, side, barOptions.priority or self.defauls.priority)
        self:Invalidate()
    end
end

function barTemplate:updateBarPlacementManual()
    local barCache = self.cache
    local frame = barCache.frame
    local options = self:getOptions()
    SidePanel:Unregister(frame)
    frame:ClearAllPoints()
    frame:SetPoint('TOP'   , UIParent, 'BOTTOM', 0            , options.top   )
    frame:SetPoint('BOTTOM', UIParent, 'BOTTOM', 0            , options.bottom)
    frame:SetPoint('LEFT'  , UIParent, 'LEFT'  , options.left , 0             )
    frame:SetPoint('RIGHT' , UIParent, 'LEFT'  , options.right, 0             )

    local oTA = self.optionsTable.args.layout.args
    oTA.top.min    = options.bottom
    oTA.bottom.max = options.top
    oTA.left.max   = options.right
    oTA.right.min  = options.left
    self:Invalidate()
end

--- Destroys the bar, this hides the frame and removes it from the layout.
function barTemplate:destroyBar()
    local barOptions = self:getOptions()
    assert(barOptions)
    local barCache = self.cache
    if not barCache then
        return
    end

    local frame = barCache.frame
    if not frame then
        return
    end

    SidePanel:Unregister(frame)
    frame:Hide()
    self:Unhook(frame,'OnSizeChanged')
    self:Unhook(frame,'OnUpdate')
end

--[==[
-- Tooltip stuff
--]==]
local GameTooltip = _G.GameTooltip
function barTemplate:OnEnter()
    local frame = self.cache.frame
    assert(frame)
    GameTooltip:SetOwner(frame, 'ANCHOR_NONE')
    local side = self:getOptions().side or self.defaults.side
    if side == 'TOP' then
        GameTooltip:SetPoint('BOTTOM', frame, side)
    elseif side == 'LEFT' then
        GameTooltip:SetPoint('RIGHT', frame, side)
    elseif side == 'BOTTOM' then
        GameTooltip:SetPoint('TOP', frame, side)
    elseif side == 'RIGHT' then
        GameTooltip:SetPoint('LEFT', frame, side)
    end
    GameTooltip:ClearLines()
    self:fillTooltip(GameTooltip)
    GameTooltip:Show()
end

function barTemplate:fillTooltip(tt)
    error("abstract function defined in barTemplate needs to be implemented")
end

local userLocale = Locales:getUserLocale()

function barTemplate:addNameToTooltip(tooltip)
    if ns.debugTooltip then
        addon:Debug(self:GetName().." in barTemplate:addNameToTooltip()")
    end
    local name = self:GetName()
    tooltip:AddLine(name)
end

function barTemplate:fillTooltipOnlyPrimary(tooltip, value, total)
    local percent = value / total * 100
    local strValue = userLocale:formatNumber(value)
    local strPercent = string.format('%.2f%%', percent)
    local strTotal = userLocale:formatNumber(total)
    tooltip:AddLine(strValue .. '/'
                 .. strTotal .. ' [' .. strPercent .. ']')
end

function barTemplate:fillTooltipWithSecond(tooltip, value, total, bonusOrCap)
    local percent = value / total * 100
    local strValue = userLocale:formatNumber(value)
    local strPercent = string.format('%.2f%%', percent)
    local strTotal = userLocale:formatNumber(total)
    local strBonusOrCap, strBonusOrCapPercent = '?', '?'
    if bonusOrCap then
        local bonusOrCapPercent = bonusOrCap / total * 100
        strBonusOrCap = userLocale:formatNumber(bonusOrCap)
        strBonusOrCapPercent = string.format('%.2f%%', bonusOrCapPercent)
    end
    tooltip:AddLine(strValue .. ' (+' .. strBonusOrCap .. ')/'
                 .. strTotal .. ' [' .. strPercent .. ' (+' .. strBonusOrCapPercent .. ')]')
end

function barTemplate:OnLeave()
    GameTooltip:Hide()
end


--- Draws the primary bar (and creates if necesary)
function barTemplate:DrawPrimaryBar(ratio)
    local barCache = self.cache
    local frame = barCache.frame
    local primaryBarTexture = barCache.primaryBarTexture
    if not primaryBarTexture then
        barCache.primaryBarTexture = barCache.primaryBarTexture or frame:CreateTexture(nil,'ARTWORK')
        primaryBarTexture = barCache.primaryBarTexture
    end
    local c = self:getOptions().priColor or self.defaults.priColor
    primaryBarTexture:SetTexture(c.r,c.g,c.b,c.a)

    local flow = self:getOptions().flow or self.defaults.flow
    local width = frame:GetRight() - frame:GetLeft()
    local height = frame:GetTop() - frame:GetBottom()
    local barWidth = ratio * width
    local barHeight = ratio * height
    primaryBarTexture:ClearAllPoints()
    if flow == 'LR' then
        primaryBarTexture:SetPoint('TOPLEFT')
        primaryBarTexture:SetPoint('BOTTOM')
        primaryBarTexture:SetWidth(max(barWidth,0.1))
    elseif flow == 'RL' then
        primaryBarTexture:SetPoint('TOPRIGHT')
        primaryBarTexture:SetPoint('BOTTOM')
        primaryBarTexture:SetWidth(max(barWidth,0.1))
    elseif flow == 'TB' then
        primaryBarTexture:SetPoint('TOPLEFT')
        primaryBarTexture:SetPoint('RIGHT')
        primaryBarTexture:SetHeight(max(barHeight,0.1))
    elseif flow == 'BT' then
        primaryBarTexture:SetPoint('BOTTOMLEFT')
        primaryBarTexture:SetPoint('RIGHT')
        primaryBarTexture:SetHeight(max(barHeight,0.1))
    else
        error("Unexpected flow type: '" .. tostring(flow) .. "'")
    end
    return primaryBarTexture
end

function barTemplate:DrawSecondaryBar(ratio)
    local barCache = self.cache
    local frame = barCache.frame
    local primaryBarTexture = barCache.primaryBarTexture
    assert(primaryBarTexture)
    local secondaryBarTexture = barCache.secondaryBarTexture
    if not secondaryBarTexture then
        barCache.secondaryBarTexture = barCache.secondaryBarTexture or frame:CreateTexture(nil,'ARTWORK')
        secondaryBarTexture = barCache.secondaryBarTexture
    end
    local c = self:getOptions().secColor or self.defaults.secColor
    secondaryBarTexture:SetTexture(c.r,c.g,c.b,c.a)

    local flow = self:getOptions().flow or self.defaults.flow
    local width = frame:GetRight() - frame:GetLeft()
    local height = frame:GetTop() - frame:GetBottom()
    local barWidth = ratio * width
    local barHeight = ratio * height
    secondaryBarTexture:ClearAllPoints()
    if flow == 'LR' then
        secondaryBarTexture:SetPoint('TOP')
        secondaryBarTexture:SetPoint('BOTTOM')
        secondaryBarTexture:SetPoint('LEFT', primaryBarTexture, 'RIGHT')
        secondaryBarTexture:SetWidth(max(barWidth,0.1))
    elseif flow == 'RL' then
        secondaryBarTexture:SetPoint('TOP')
        secondaryBarTexture:SetPoint('BOTTOM')
        secondaryBarTexture:SetPoint('RIGHT', primaryBarTexture, 'LEFT')
        secondaryBarTexture:SetWidth(max(barWidth,0.1))
    elseif flow == 'TB' then
        secondaryBarTexture:SetPoint('LEFT')
        secondaryBarTexture:SetPoint('RIGHT')
        secondaryBarTexture:SetPoint('TOP', primaryBarTexture, 'BOTTOM')
        secondaryBarTexture:SetHeight(max(barHeight,0.1))
    elseif flow == 'BT' then
        secondaryBarTexture:SetPoint('LEFT')
        secondaryBarTexture:SetPoint('RIGHT')
        secondaryBarTexture:SetPoint('BOTTOM', primaryBarTexture, 'TOP')
        secondaryBarTexture:SetHeight(max(barHeight,0.1))
    else
        error("Unexpected flow type: '" .. tostring(flow) .. "'")
    end
    return secondaryBarTexture
end